# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import asyncio

from mautrix.client import Client
from mautrix.bridge.commands import HelpSection, command_handler
from mautrix.bridge import custom_puppet as cpu

from maufbapi import AndroidState, AndroidAPI
from maufbapi.http import TwoFactorRequired, OAuthException, IncorrectPassword

from .. import puppet as pu
from .typehint import CommandEvent

SECTION_AUTH = HelpSection("Authentication", 10, "")


async def check_approved_login(state: AndroidState, api: AndroidAPI, evt: CommandEvent) -> None:
    while evt.sender.command_status and evt.sender.command_status["action"] == "Login":
        await asyncio.sleep(5)
        try:
            was_approved = await api.check_approved_machine()
        except Exception as e:
            evt.log.exception("Error checking if login was approved from another device")
            await evt.reply(f"Error checking if login was approved from another device: {e}")
            break
        if was_approved:
            prev_cmd_status = evt.sender.command_status
            evt.sender.command_status = None
            try:
                await api.login_approved()
            except TwoFactorRequired:
                await evt.reply("Login approved from another device, but Facebook decided that "
                                "you need to enter the 2FA code anyway.")
                evt.sender.command_status = prev_cmd_status
                return
            await evt.sender.on_logged_in(state)
            await evt.reply("Login successfully approved from another device")
            break


@command_handler(needs_auth=False, management_only=True,
                 help_section=SECTION_AUTH, help_text="Log in to Facebook",
                 help_args="<_email_> <_password_>")
async def login(evt: CommandEvent) -> None:
    if len(evt.args) < 2:
        await evt.reply("Usage: `$cmdprefix+sp login <email> <password>`")
        return

    email, password = evt.args[0], " ".join(evt.args[1:])
    await evt.az.intent.redact(evt.room_id, evt.event_id)

    if evt.sender.client:
        await evt.reply("You're already logged in")
        return

    state = AndroidState()
    state.generate(evt.sender.mxid)
    api = AndroidAPI(state, log=evt.sender.log.getChild("login-api"))
    try:
        await api.mobile_config_sessionless()
        await api.login(email, password)
        await evt.sender.on_logged_in(state)
        await evt.reply("Successfully logged in")
    except TwoFactorRequired:
        await evt.reply("You have two-factor authentication turned on. Please either send the code"
                        " from SMS or your authenticator app here, or approve the login from"
                        " another device logged into Messenger.")
        checker_task = asyncio.create_task(check_approved_login(state, api, evt))
        evt.sender.command_status = {
            "action": "Login",
            "room_id": evt.room_id,
            "next": enter_2fa_code,
            "state": state,
            "api": api,
            "email": email,
            "checker_task": checker_task,
        }
    except OAuthException as e:
        await evt.reply(f"Error from Messenger:\n\n> {e}")
    except Exception as e:
        evt.sender.command_status = None
        await evt.reply(f"Failed to log in: {e}")
        evt.log.exception("Failed to log in")


async def enter_2fa_code(evt: CommandEvent) -> None:
    checker_task: asyncio.Task = evt.sender.command_status["checker_task"]
    checker_task.cancel()
    state: AndroidState = evt.sender.command_status["state"]
    api: AndroidAPI = evt.sender.command_status["api"]
    email: str = evt.sender.command_status["email"]
    try:
        await api.login_2fa(email, "".join(evt.args).strip())
        await evt.sender.on_logged_in(state)
        await evt.reply("Successfully logged in")
        evt.sender.command_status = None
    except IncorrectPassword:
        await evt.reply("Incorrect two-factor authentication code. Pleaase try again.")
    except OAuthException as e:
        await evt.reply(f"Error from Messenger:\n\n> {e}")
        evt.sender.command_status = None
    except Exception as e:
        evt.sender.command_status = None
        await evt.reply(f"Failed to log in: {e}")
        evt.log.exception("Failed to log in")


@command_handler(needs_auth=True, help_section=SECTION_AUTH, help_text="Log out of Facebook")
async def logout(evt: CommandEvent) -> None:
    puppet = await pu.Puppet.get_by_fbid(evt.sender.fbid)
    await evt.sender.logout()
    if puppet.is_real_user:
        await puppet.switch_mxid(None, None)
    await evt.reply("Successfully logged out")


@command_handler(needs_auth=True, management_only=True, help_args="<_access token_>",
                 help_section=SECTION_AUTH, help_text="Replace your Facebook Messenger account's "
                                                      "Matrix puppet with your Matrix account")
async def login_matrix(evt: CommandEvent) -> None:
    puppet = await pu.Puppet.get_by_fbid(evt.sender.fbid)
    _, homeserver = Client.parse_mxid(evt.sender.mxid)
    if homeserver != pu.Puppet.hs_domain:
        await evt.reply("You can't log in with an account on a different homeserver")
        return
    try:
        await puppet.switch_mxid(" ".join(evt.args), evt.sender.mxid)
        await evt.reply("Successfully replaced your Facebook Messenger account's "
                        "Matrix puppet with your Matrix account.")
    except cpu.OnlyLoginSelf:
        await evt.reply("You may only log in with your own Matrix account")
    except cpu.InvalidAccessToken:
        await evt.reply("Invalid access token")


@command_handler(needs_auth=True, management_only=True, help_section=SECTION_AUTH,
                 help_text="Revert your Facebook Messenger account's Matrix puppet to the original")
async def logout_matrix(evt: CommandEvent) -> None:
    puppet = await pu.Puppet.get_by_fbid(evt.sender.fbid)
    if not puppet.is_real_user:
        await evt.reply("You're not logged in with your Matrix account")
        return
    await puppet.switch_mxid(None, None)
    await evt.reply("Restored the original puppet for your Facebook Messenger account")



from .queries import (GraphQLQuery, GraphQLMutation, MoreMessagesQuery, ThreadListQuery,
                      ThreadNameMutation, ThreadNameMutationSource, FetchStickersWithPreviewsQuery,
                      MessageUndoSend, ReactionAction, MessageReactionMutation, ThreadMessageID,
                      DownloadImageFragment, SubsequentMediaQuery, FbIdToCursorQuery, ThreadQuery,
                      FileAttachmentUrlQuery, SearchEntitiesNamedQuery)
from .responses import (ParticipantID, ReadReceipt, ReadReceiptList, Picture, StructuredNameChunk,
                        StructuredName, FriendCount, MinimalParticipant, ParticipantNode,
                        Participant, ParticipantList, MessageSender, MessageRange, MessagePowerUp,
                        MessageText, Reaction, Dimensions, Attachment, Sticker, MinimalMessage,
                        ReplyStatus, Reply, Message, PageInfo, MessageList, ThreadKey, Thread,
                        ThreadListResponse, StickerPreviewResponse, MinimalSticker, ImageFragment,
                        StickerPackMeta, MessageUnsendResponse, AttachmentType, ExtensibleText,
                        ExtensibleAttachment, StoryAttachment, SubsequentMediaNode, Coordinates,
                        SubsequentMediaResponse, FileAttachmentURLResponse, FileAttachmentWithURL,
                        StoryTarget, OwnInfo, SearchEntitiesResponse, SearchResults, SearchResult,
                        ThreadQueryResponse)



import subprocess
import shutil
import os

from . import __version__

cmd_env = {
    "PATH": os.environ["PATH"],
    "HOME": os.environ["HOME"],
    "LANG": "C",
    "LC_ALL": "C",
}


def run(cmd):
    return subprocess.check_output(cmd, stderr=subprocess.DEVNULL, env=cmd_env)

if os.path.exists(".git") and shutil.which("git"):
    try:
        git_revision = run(["git", "rev-parse", "HEAD"]).strip().decode("ascii")
        git_revision_url = f"https://github.com/tulir/mautrix-facebook/commit/{git_revision}"
        git_revision = git_revision[:8]
    except (subprocess.SubprocessError, OSError):
        git_revision = "unknown"
        git_revision_url = None

    try:
        git_tag = run(["git", "describe", "--exact-match", "--tags"]).strip().decode("ascii")
    except (subprocess.SubprocessError, OSError):
        git_tag = None
else:
    git_revision = "unknown"
    git_revision_url = None
    git_tag = None

git_tag_url = (f"https://github.com/tulir/mautrix-facebook/releases/tag/{git_tag}"
               if git_tag else None)

if git_tag and __version__ == git_tag[1:].replace("-", ""):
    version = __version__
    linkified_version = f"[{version}]({git_tag_url})"
else:
    if not __version__.endswith("+dev"):
        __version__ += "+dev"
    version = f"{__version__}.{git_revision}"
    if git_revision_url:
        linkified_version = f"{__version__}.[{git_revision}]({git_revision_url})"
    else:
        linkified_version = version



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, List, TYPE_CHECKING, ClassVar

from asyncpg import Record
from attr import dataclass

from mautrix.types import UserID, RoomID
from mautrix.util.async_db import Database
from maufbapi import AndroidState

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class User:
    db: ClassVar[Database] = fake_db

    mxid: UserID
    fbid: Optional[int]
    state: Optional[AndroidState]
    notice_room: Optional[RoomID]

    @property
    def _state_json(self) -> Optional[str]:
        return self.state.json() if self.state else None

    @classmethod
    def _from_row(cls, row: Optional[Record]) -> Optional['User']:
        if row is None:
            return None
        data = {**row}
        state = data.pop("state", None)
        return cls(**data, state=AndroidState.parse_json(state) if state else None)

    @classmethod
    async def all_logged_in(cls) -> List['User']:
        rows = await cls.db.fetch('SELECT mxid, fbid, state, notice_room FROM "user" '
                                  "WHERE fbid<>0 AND state IS NOT NULL")
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def get_by_fbid(cls, fbid: int) -> Optional['User']:
        q = 'SELECT mxid, fbid, state, notice_room FROM "user" WHERE fbid=$1'
        row = await cls.db.fetchrow(q, fbid)
        return cls._from_row(row)

    @classmethod
    async def get_by_mxid(cls, mxid: UserID) -> Optional['User']:
        q = 'SELECT mxid, fbid, state, notice_room FROM "user" WHERE mxid=$1'
        row = await cls.db.fetchrow(q, mxid)
        return cls._from_row(row)

    async def insert(self) -> None:
        q = 'INSERT INTO "user" (mxid, fbid, state, notice_room) VALUES ($1, $2, $3, $4)'
        await self.db.execute(q, self.mxid, self.fbid, self._state_json, self.notice_room)

    async def delete(self) -> None:
        await self.db.execute('DELETE FROM "user" WHERE mxid=$1', self.mxid)

    async def save(self) -> None:
        await self.db.execute('UPDATE "user" SET fbid=$2, state=$3, notice_room=$4 WHERE mxid=$1',
                              self.mxid, self.fbid, self._state_json, self.notice_room)



from mautrix.util.async_db import Database

from .upgrade import upgrade_table
from .message import Message
from .reaction import Reaction
from .portal import Portal, ThreadType
from .puppet import Puppet
from .user import User
from .user_portal import UserPortal
from .user_contact import UserContact


def init(db: Database) -> None:
    for table in (Portal, Message, Reaction, User, Puppet, UserPortal, UserContact):
        table.db = db


__all__ = ["upgrade_table", "init", "Message", "Reaction", "Portal", "ThreadType", "Puppet",
           "User", "UserPortal", "UserContact"]



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, List, ClassVar
from abc import ABC

from attr import dataclass
import attr

from mautrix.types import Serializable, SerializableAttrs, SerializableEnum
from ..common import ThreadFolder


class GraphQLQuery(ABC, Serializable):
    caller_class: ClassVar[str] = "graphservice"
    doc_id: ClassVar[int]


class GraphQLMutation(GraphQLQuery, ABC):
    pass


@dataclass
class NTContext(SerializableAttrs['ThreadListQueryNTContext']):
    styles_id: str = "7d328425a4dfa3aa76b1310fa8dc30bf"
    using_white_navbar: bool = True
    pixel_ratio: int = 3


@dataclass
class ThreadQuery(GraphQLQuery, SerializableAttrs['ThreadQuery']):
    doc_id: ClassVar[int] = 5487678687924830

    thread_ids: List[str]
    msg_count: int = 20

    blur: int = 0

    nt_context: NTContext = attr.ib(factory=lambda: NTContext())
    include_full_user_info: str = "true"
    include_message_info: str = "true"
    include_booking_requests: bool = True

    full_screen_width: int = 4096
    full_screen_height: int = 4096
    large_preview_width: int = 1500
    large_preview_height: int = 750
    medium_preview_width: int = 962
    medium_preview_height: int = 481
    small_preview_width: int = 716
    small_preview_height: int = 358
    profile_pic_large_size: int = 880
    profile_pic_small_size: int = 138



@dataclass
class ThreadListQuery(GraphQLQuery, SerializableAttrs['ThreadListQuery']):
    doc_id: ClassVar[int] = 3562683343826563

    msg_count: int = 20
    thread_count: int = 20
    include_thread_info: str = "true"
    include_message_info: str = "true"
    fetch_users_separately: str = "false"
    filter_to_groups: str = "false"
    include_booking_requests: bool = True

    nt_context: NTContext = attr.ib(factory=lambda: NTContext())
    folder_tag: Optional[List[ThreadFolder]] = None

    theme_icon_size_small: int = 66
    reaction_static_asset_size_small: int = 39
    profile_pic_medium_size: int = 220
    profile_pic_large_size: int = 880
    profile_pic_small_size: int = 138
    theme_background_size: int = 2048
    theme_icon_size_large: int = 138


@dataclass
class MoreMessagesQuery(GraphQLQuery, SerializableAttrs['MoreMessagesQuery']):
    doc_id: ClassVar[int] = 3447218621980314

    before_time_ms: str
    thread_id: str
    msg_count: int = 20
    blur: int = 0

    nt_context: NTContext = attr.ib(factory=lambda: NTContext())

    full_screen_width: int = 4096
    full_screen_height: int = 4096
    large_preview_width: int = 1500
    large_preview_height: int = 750
    medium_preview_width: int = 962
    medium_preview_height: int = 481
    small_preview_width: int = 716
    small_preview_height: int = 358


class ThreadNameMutationSource(SerializableEnum):
    SETTINGS = "SETTINGS"


@dataclass
class ThreadNameMutation(GraphQLMutation, SerializableAttrs['ThreadNameMutation']):
    doc_id: ClassVar[int] = 3090707060965997

    new_thread_name: str
    thread_id: str
    client_mutation_id: str
    actor_id: str
    source: ThreadNameMutationSource = ThreadNameMutationSource.SETTINGS


@dataclass
class FetchStickersWithPreviewsQuery(GraphQLQuery,
                                     SerializableAttrs['FetchStickersWithPreviewsQuery']):
    doc_id: ClassVar[int] = 3154119451330002
    caller_class: ClassVar[str] = "com.facebook.messaging.sync.delta.NewMessageHandlerHelper"

    sticker_ids: List[str]
    preview_size: int = 165
    animated_media_type: str = "image/webp"
    media_type: str = "image/webp"
    scaling_factor: str = "2.75"
    sticker_labels_enabled: bool = False
    sticker_state_enabled: bool = False


@dataclass
class MessageUndoSend(GraphQLMutation, SerializableAttrs['MessageUndoSend']):
    doc_id: ClassVar[int] = 1015037405287590

    message_id: str
    client_mutation_id: str
    actor_id: str


class ReactionAction(SerializableEnum):
    ADD = "ADD_REACTION"
    REMOVE = "REMOVE_REACTION"


@dataclass
class MessageReactionMutation(GraphQLMutation, SerializableAttrs['MessageReactionMutation']):
    doc_id: ClassVar[int] = 1415891828475683

    message_id: str
    client_mutation_id: str
    actor_id: str
    action: ReactionAction
    reaction: Optional[str] = None


@dataclass
class DownloadImageFragment(GraphQLQuery, SerializableAttrs['DownloadImageFragment']):
    doc_id: ClassVar[int] = 3063616537053520

    fbid: str
    img_size: str = "0"


@dataclass
class FbIdToCursorQuery(GraphQLQuery, SerializableAttrs['FbIdToCursorQuery']):
    doc_id: ClassVar[int] = 2015407048575350

    fbid: str
    thread_id: str


@dataclass
class SubsequentMediaQuery(GraphQLQuery, SerializableAttrs['SubsequentMediaQuery']):
    doc_id: ClassVar[int] = 2948398158550055

    thread_id: str
    cursor_id: Optional[str] = None
    fetch_size: int = 99
    thumbnail_size: int = 540
    height: int = 2088
    width: int = 1080


@dataclass(frozen=True, eq=True)
class ThreadMessageID(SerializableAttrs['ThreadMessageID']):
    thread_id: str
    message_id: str


@dataclass
class FileAttachmentUrlQuery(GraphQLQuery, SerializableAttrs['FileAttachmentUrlQuery']):
    doc_id: ClassVar[int] = 3200288700012393

    thread_msg_id: ThreadMessageID


@dataclass
class SearchEntitiesNamedQuery(GraphQLQuery, SerializableAttrs['SearchEntitiesNamedQuery']):
    doc_id: ClassVar[int] = 3414226858659179

    search_query: str
    session_id: Optional[str] = None

    results_limit: int = 20
    num_users_query: int = 20
    num_group_threads_query: int = 20
    num_pages_query: int = 6

    unified_config: str = "DOUBLE_SERVER_QUERY_PRIMARY"
    search_surface: str = "UNIVERSAL_ALL"
    user_types: List[str] = ["CONTACT", "NON_FRIEND_NON_CONTACT"]
    entity_types: List[str] = ["user", "group_thread", "page", "game", "matched_message_thread"]
    include_pages: bool = True
    include_games: bool = True

    profile_pic_large_size: int = 880
    profile_pic_medium_size: int = 220
    profile_pic_small_size: int = 138



"""Switch to mautrix-python crypto

Revision ID: 905a91536747
Revises: 244273a86d81
Create Date: 2020-07-08 19:02:19.909677

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '905a91536747'
down_revision = '244273a86d81'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('nio_account')
    op.drop_table('nio_device_key')
    op.drop_table('nio_outgoing_key_request')
    op.drop_table('nio_olm_session')
    op.drop_table('nio_megolm_inbound_session')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('nio_megolm_inbound_session',
    sa.Column('session_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('sender_key', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('fp_key', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('room_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('session', postgresql.BYTEA(), autoincrement=False, nullable=False),
    sa.Column('forwarded_chains', postgresql.BYTEA(), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('session_id', name='nio_megolm_inbound_session_pkey')
    )
    op.create_table('nio_olm_session',
    sa.Column('session_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('sender_key', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('session', postgresql.BYTEA(), autoincrement=False, nullable=False),
    sa.Column('created_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=False),
    sa.Column('last_used', postgresql.TIMESTAMP(), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('session_id', name='nio_olm_session_pkey')
    )
    op.create_table('nio_outgoing_key_request',
    sa.Column('request_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('session_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('room_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('algorithm', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('request_id', name='nio_outgoing_key_request_pkey')
    )
    op.create_table('nio_device_key',
    sa.Column('user_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('device_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('display_name', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('deleted', sa.BOOLEAN(), autoincrement=False, nullable=False),
    sa.Column('keys', postgresql.BYTEA(), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('user_id', 'device_id', name='nio_device_key_pkey')
    )
    op.create_table('nio_account',
    sa.Column('user_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('device_id', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('shared', sa.BOOLEAN(), autoincrement=False, nullable=False),
    sa.Column('sync_token', sa.TEXT(), autoincrement=False, nullable=False),
    sa.Column('account', postgresql.BYTEA(), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('user_id', 'device_id', name='nio_account_pkey')
    )
    # ### end Alembic commands ###



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.types import ExtensibleEnum, SerializableEnum


class MessageUnsendability(ExtensibleEnum):
    DENY_FOR_NON_SENDER = "deny_for_non_sender"
    DENY_LOG_MESSAGE = "deny_log_message"
    DENY_TOMBSTONE_MESSAGE = "deny_tombstone_message"
    DENY_IF_PAGE_THREAD = "deny_if_page_thread"
    DENY_IF_CANNOT_REPLY_TO_VIEWER_THREAD = "deny_if_cannot_reply_to_viewer_thread"
    DENY_IF_MARKETPLACE_THREAD = "deny_if_marketplace_thread"
    CAN_UNSEND = "can_unsend"


class ThreadFolder(SerializableEnum):
    INBOX = "INBOX"
    PENDING = "PENDING"
    ARCHIVED = "ARCHIVED"
    OTHER = "OTHER"



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List

from attr import dataclass

from maufbapi.thrift import TType, ThriftObject, field, autospec


@autospec
@dataclass(kw_only=True)
class RealtimeClientInfo(ThriftObject):
    user_id: int = field(TType.I64)
    user_agent: str
    client_capabilities: int = field(TType.I64)
    endpoint_capabilities: int = field(TType.I64)
    # 0 = no zlib?, 1 = always zlib, 2 = optional zlib
    publish_format: int = field(TType.I32)
    no_automatic_foreground: bool
    make_user_available_in_foreground: bool
    device_id: str
    is_initially_foreground: bool
    network_type: int = field(TType.I32)
    network_subtype: int = field(TType.I32)
    client_mqtt_session_id: int = field(TType.I64)
    client_ip_address: str = None
    subscribe_topics: List[int] = field(TType.LIST, item_type=TType.I32)
    client_type: str
    app_id: int = field(TType.I64)
    override_nectar_logging: bool = None
    connect_token_hash: str = None
    region_preference: str
    device_secret: str
    client_stack: int = field(TType.BYTE)
    fbns_connection_key: int = field(TType.I64, default=None)
    fbns_connection_secret: str = None
    fbns_device_id: str = None
    fbns_device_secret: str = None
    another_unknown: int = field(TType.I64, default=None)
    yet_another_unknown: int = field(TType.I32, default=None)


@autospec
@dataclass(kw_only=True)
class RealtimeConfig(ThriftObject):
    client_identifier: str
    will_topic: str = None
    will_message: str = None
    client_info: RealtimeClientInfo
    password: str
    get_diffs_request: List[str] = None
    zero_rating_token_hash: str = None
    # mysterious_struct_list: List[Any] = field(TType.LIST, TType.STRUCT, factory=lambda: [])
    app_specific_info: Dict[str, str] = field(index=9)


@autospec
@dataclass(kw_only=True)
class ForegroundStateConfig(ThriftObject):
    in_foreground_app: bool
    in_foreground_device: bool
    keep_alive_timeout: int = field(TType.I32)
    subscribe_topics: List[str]
    subscribe_generic_topics: List[str]
    unsubscribe_topics: List[str]
    unsubscribe_generic_topics: List[str]
    request_id: int = field(TType.I64)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Tuple, Union, Optional, Dict, Type, NamedTuple, Any, TypeVar
import sys

import attr

from .type import TType, RecursiveType
from .write import ThriftWriter
from .read import ThriftReader

T = TypeVar('T')


class ThriftField(NamedTuple):
    name: str
    rtype: RecursiveType


class ThriftObject:
    thrift_spec: Dict[int, ThriftField]

    def to_thrift(self) -> bytes:
        buf = ThriftWriter()
        buf.write_struct(self)
        return buf.getvalue()

    @classmethod
    def from_thrift(cls: Type[T], data: bytes) -> T:
        return ThriftReader(data).read_struct(cls)


TYPE_META = "net.maunium.thrift.type"
INDEX_META = "net.maunium.thrift.index"

if sys.version_info >= (3, 7):
    def _get_type_class(typ):
        try:
            return typ.__origin__
        except AttributeError:
            return None
else:
    def _get_type_class(typ):
        try:
            return typ.__extra__
        except AttributeError:
            return None


def _guess_type(python_type, name: str) -> RecursiveType:
    if python_type == str or python_type == bytes:
        return RecursiveType(TType.BINARY, python_type=python_type)
    elif python_type == bool:
        return RecursiveType(TType.BOOL, python_type=python_type)
    elif python_type == int:
        raise ValueError(f"Ambiguous integer field {name}")
    elif python_type == float:
        return RecursiveType(TType.DOUBLE, python_type=python_type)
    elif attr.has(python_type):
        return RecursiveType(TType.STRUCT, python_type=python_type)

    type_class = _get_type_class(python_type)
    args = getattr(python_type, "__args__", None)
    if type_class == list:
        return RecursiveType(TType.LIST, item_type=_guess_type(args[0], f"{name} item"),
                             python_type=list)
    elif type_class == dict:
        return RecursiveType(TType.MAP, key_type=_guess_type(args[0], f"{name} key"),
                             value_type=_guess_type(args[1], f"{name} value"), python_type=dict)
    elif type_class == set:
        return RecursiveType(TType.SET, item_type=_guess_type(args[0], f"{name} item"),
                             python_type=set)

    raise ValueError(f"Unknown type {python_type} for {name}")


def autospec(clazz: Any) -> Type[ThriftObject]:
    """
    Automatically generate a thrift_spec dict based on attrs metadata.

    Args:
        clazz: The class to decorate.

    Returns:
        The class given as a parameter.
    """
    clazz.thrift_spec = {}
    index = 0
    for field in attr.fields(clazz):
        try:
            field_type = field.metadata[TYPE_META]._replace(python_type=field.type)
        except KeyError:
            field_type = _guess_type(field.type, field.name)
        field_meta = ThriftField(field.name, field_type)
        try:
            index = field.metadata[INDEX_META]
        except KeyError:
            index += 1
        clazz.thrift_spec[index] = field_meta
    return clazz


RTType = Union[None, TType, RecursiveType]


def _rttype_to_rtype(typ: RTType, item_type: RTType = None, key_type: RTType = None,
                     value_type: RTType = None) -> Optional[RecursiveType]:
    if not typ:
        return None
    elif isinstance(typ, RecursiveType):
        return typ
    return RecursiveType(typ, item_type=_rttype_to_rtype(item_type),
                         key_type=_rttype_to_rtype(key_type),
                         value_type=_rttype_to_rtype(value_type))


def field(thrift_type: RTType = None, index: Optional[int] = None, item_type: RTType = None,
          key_type: RTType = None, value_type: RTType = None, **kwargs) -> attr.Attribute:
    """
    Specify an explicit type for the :meth:`autospec` decorator.

    Args:
        thrift_type: The thrift type to use for the field.
        index: Override value for the Thrift field index.
        item_type: Type of list/set items.
        key_type: Type of map keys.
        value_type: Type of map values.
        **kwargs: Other parameters to pass to :meth:`attr.ib`.

    Returns:
        The result of :meth:`attr.ib`
    """
    meta = kwargs.setdefault("metadata", {})
    if thrift_type is not None:
        meta[TYPE_META] = _rttype_to_rtype(thrift_type, item_type, key_type, value_type)
    if index is not None:
        meta[INDEX_META] = index
    return attr.ib(**kwargs)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Dict
from uuid import UUID
import random
import string

import attr
from attr import dataclass
from mautrix.types import SerializableAttrs


@dataclass
class AndroidApplication(SerializableAttrs['AndroidApplication']):
    name: str = "Orca-Android"
    version: str = "294.0.0.24.129"
    id: str = "com.facebook.orca"
    locale: str = "en_US"
    build: int = 263695262
    version_id: int = 3402226163209239

    client_id = "256002347743983"
    client_secret = "374e60f8b9bb6b8cbb30f78030438895"

    @property
    def access_token(self) -> str:
        return f"{self.client_id}|{self.client_secret}"


@dataclass
class AndroidDevice(SerializableAttrs['AndroidDevice']):
    manufacturer: str = "Google"
    builder: str = "google"
    name: str = "Pixel 3"
    software: str = "10"
    architecture: str = "arm64-v8a:null"
    dimensions: str = "{density=2.75,width=1080,height=2028}"
    user_agent: str = "Dalvik/2.1.0 (Linux; U; Android 10; Pixel 3 Build/QQ3A.200605.001)"

    connection_type: str = "WIFI"
    connection_quality: str = "EXCELLENT"
    language: str = "en_US"
    country_code: str = "US"

    uuid: Optional[str] = None
    adid: Optional[str] = None

    device_group: Optional[str] = None  # 7761


@dataclass
class AndroidCarrier(SerializableAttrs['AndroidCarrier']):
    name: str = "Verizon"
    hni: int = 311390


@dataclass
class AndroidSession(SerializableAttrs['AndroidSession']):
    access_token: Optional[str] = None
    uid: Optional[int] = None
    password_encryption_pubkey: Optional[str] = None
    password_encryption_key_id: Optional[int] = None
    machine_id: Optional[str] = None
    transient_auth_token: Optional[str] = None
    login_first_factor: Optional[str] = None
    region_hint: str = "ODN"


@dataclass
class AndroidState(SerializableAttrs['AndroidState']):
    application: AndroidApplication = attr.ib(factory=lambda: AndroidApplication())
    device: AndroidDevice = attr.ib(factory=lambda: AndroidDevice())
    carrier: AndroidCarrier = attr.ib(factory=lambda: AndroidCarrier())
    session: AndroidSession = attr.ib(factory=lambda: AndroidSession())

    def generate(self, seed: str) -> None:
        rand = random.Random(seed)
        self.device.adid = "".join(rand.choices(string.hexdigits, k=16))
        self.device.uuid = str(UUID(int=rand.getrandbits(128), version=4))
        # TODO randomize carrier and device model

    @property
    def _ua_parts(self) -> Dict[str, str]:
        return {
            "FBAN": self.application.name,
            "FBAV": self.application.version,
            "FBPN": self.application.id,
            "FBLC": self.device.language,
            "FBBV": str(self.application.build),
            "FBCR": self.carrier.name,
            "FBMF": self.device.manufacturer,
            "FBBD": self.device.builder,
            "FBDV": self.device.name,
            "FBSV": self.device.software,
            "FBCA": self.device.architecture,
            "FBDM": self.device.dimensions,
            "FB_FW": "1",
        }

    @property
    def user_agent_meta(self) -> str:
        ua_meta = ";".join(f"{key}/{value}" for key, value in self._ua_parts.items())
        return f"[{ua_meta};]"

    @property
    def user_agent(self) -> str:
        return f"{self.device.user_agent} {self.user_agent_meta}"



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from datetime import timezone
from enum import Enum as EnumType

from sqlalchemy import (Column, Text, SmallInteger, Boolean, Enum, PickleType, UniqueConstraint,
                        ForeignKey, ForeignKeyConstraint, false, types)

from mautrix.util.db import Base


class UTCDateTime(types.TypeDecorator):
    impl = types.DateTime

    def process_bind_param(self, value, dialect):
        if value is not None:
            if value.tzinfo is None:
                value = value.replace(tzinfo=timezone.utc)
            elif value.tzinfo != timezone.utc:
                value = value.astimezone(timezone.utc)

        return value

    def process_result_value(self, value, dialect):
        if value is not None and value.tzinfo is None:
            return value.replace(tzinfo=timezone.utc)
        else:
            return value


class ThreadType(EnumType):
    USER = 1
    GROUP = 2
    PAGE = 3
    UNKNOWN = 4


class Message(Base):
    __tablename__ = "message"

    mxid = Column(Text, nullable=False)
    mx_room = Column(Text, nullable=False)
    fbid = Column(Text, primary_key=True)
    fb_chat = Column(Text, nullable=True)
    fb_receiver = Column(Text, primary_key=True)
    index = Column(SmallInteger, primary_key=True, default=0)
    date = Column(UTCDateTime(timezone=True), nullable=True)

    __table_args__ = (UniqueConstraint("mxid", "mx_room", name="_mx_id_room"),)


class Portal(Base):
    __tablename__ = "portal"

    fbid = Column(Text, primary_key=True)
    fb_receiver = Column(Text, primary_key=True)
    fb_type = Column(Enum(ThreadType), nullable=False)

    mxid = Column(Text, unique=True, nullable=True)
    avatar_url = Column(Text, nullable=True)
    encrypted = Column(Boolean, nullable=False, server_default=false())

    name = Column(Text, nullable=True)
    photo_id = Column(Text, nullable=True)


class Puppet(Base):
    __tablename__ = "puppet"

    fbid = Column(Text, primary_key=True)
    name = Column(Text, nullable=True)
    name_set = Column(Boolean, nullable=False, server_default=false())
    photo_id = Column(Text, nullable=True)
    avatar_set = Column(Boolean, nullable=False, server_default=false())
    matrix_registered = Column(Boolean, nullable=False, server_default=false())

    custom_mxid = Column(Text, nullable=True)
    access_token = Column(Text, nullable=True)
    next_batch = Column(Text, nullable=True)
    base_url = Column(Text, nullable=True)


class Reaction(Base):
    __tablename__ = "reaction"

    mxid = Column(Text, nullable=False)
    mx_room = Column(Text, nullable=False)
    fb_msgid = Column(Text, primary_key=True)
    fb_receiver = Column(Text, primary_key=True)
    fb_sender = Column(Text, primary_key=True)
    reaction = Column(Text, nullable=False)

    __table_args__ = (UniqueConstraint("mxid", "mx_room", name="_mx_react_id_room"),)


class User(Base):
    __tablename__ = "user"

    mxid = Column(Text, primary_key=True)
    session = Column(PickleType, nullable=True)
    fbid = Column(Text, nullable=True)
    notice_room = Column(Text, nullable=True)
    user_agent = Column(Text, nullable=True)
    fb_domain = Column(Text, nullable=False, server_default="messenger.com")


class Contact(Base):
    __tablename__ = "contact"

    user = Column(Text, primary_key=True)
    contact = Column(Text, ForeignKey("puppet.fbid"), primary_key=True)
    in_community = Column(Boolean, nullable=False, server_default=false())


class UserPortal(Base):
    __tablename__ = "user_portal"

    user = Column(Text, primary_key=True)
    portal = Column(Text, primary_key=True)
    portal_receiver = Column(Text, primary_key=True)
    in_community = Column(Boolean, nullable=False, server_default=false())

    __table_args__ = (ForeignKeyConstraint(("portal", "portal_receiver"),
                                           ("portal.fbid", "portal.fb_receiver"),
                                           onupdate="CASCADE", ondelete="CASCADE"),)



import setuptools
import glob

from mautrix_facebook.get_version import git_tag, git_revision, version, linkified_version

try:
    long_desc = open("README.md").read()
except IOError:
    long_desc = "Failed to read README.md"

with open("requirements.txt") as reqs:
    install_requires = reqs.read().splitlines()

with open("optional-requirements.txt") as reqs:
    extras_require = {}
    current = []
    for line in reqs.read().splitlines():
        if line.startswith("#/"):
            extras_require[line[2:]] = current = []
        elif not line or line.startswith("#"):
            continue
        else:
            current.append(line)

extras_require["all"] = list({dep for deps in extras_require.values() for dep in deps})

with open("mautrix_facebook/version.py", "w") as version_file:
    version_file.write(f"""# Generated in setup.py

git_tag = {git_tag!r}
git_revision = {git_revision!r}
version = {version!r}
linkified_version = {linkified_version!r}
""")

setuptools.setup(
    name="mautrix-facebook",
    version=version,
    url="https://github.com/tulir/mautrix-facebook",

    author="Tulir Asokan",
    author_email="tulir@maunium.net",

    description="A Matrix-Facebook Messenger puppeting bridge.",
    long_description=long_desc,
    long_description_content_type="text/markdown",

    packages=setuptools.find_packages(),

    install_requires=install_requires,
    extras_require=extras_require,
    python_requires="~=3.7",

    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Topic :: Communications :: Chat",
        "Framework :: AsyncIO",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    package_data={
        "mautrix_facebook": [
#            "web/static/*",
            "example-config.yaml",
        ],
        "maufbapi.mqtt": [
            "topics.json",
        ],
    },
    data_files=[
        (".", ["alembic.ini", "mautrix_facebook/example-config.yaml"]),
        ("alembic", ["alembic/env.py"]),
        ("alembic/versions", glob.glob("alembic/versions/*.py"))
    ],
)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import asyncio

from mautrix.bridge.commands import HelpSection, command_handler
from maufbapi.types import graphql

from .. import puppet as pu, user as u
from .typehint import CommandEvent

SECTION_MISC = HelpSection("Miscellaneous", 40, "")


async def _get_search_result_puppet(source: 'u.User', node: graphql.Participant) -> 'pu.Puppet':
    puppet = await pu.Puppet.get_by_fbid(node.id)
    if not puppet.name_set:
        await puppet.update_info(source, node)
    return puppet


@command_handler(needs_auth=True, management_only=False,
                 help_section=SECTION_MISC, help_text="Search for a Facebook user",
                 help_args="<_search query_>")
async def search(evt: CommandEvent) -> None:
    resp = await evt.sender.client.search(" ".join(evt.args))
    puppets = await asyncio.gather(*[_get_search_result_puppet(evt.sender, edge.node)
                                     for edge in resp.search_results.edges
                                     if isinstance(edge.node, graphql.Participant)])
    results = "".join(f"* [{puppet.name}](https://matrix.to/#/{puppet.default_mxid})\n"
                      for puppet in puppets)
    if results:
        await evt.reply(f"Search results:\n\n{results}")
    else:
        await evt.reply("No results :(")



"""Add date and chat ID for messages

Revision ID: 76cb89f6b352
Revises: a26da6af5cff
Create Date: 2020-06-02 14:44:13.752450

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = "76cb89f6b352"
down_revision = "a26da6af5cff"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("message") as batch_op:
        batch_op.add_column(sa.Column("date", sa.DateTime(timezone=True), nullable=True))
        batch_op.add_column(sa.Column("fb_chat", sa.String(127), nullable=True))
        batch_op.alter_column("mx_room", existing_type=sa.String(255), nullable=False)
        batch_op.alter_column("mxid", existing_type=sa.String(255), nullable=False)


def downgrade():
    with op.batch_alter_table("message") as batch_op:
        batch_op.alter_column("mxid", existing_type=sa.String(255), nullable=True)
        batch_op.alter_column("mx_room", existing_type=sa.String(255), nullable=True)
        batch_op.drop_column("fb_chat")
        batch_op.drop_column("date")



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Any, TypeVar, Type, Tuple, TYPE_CHECKING
import struct
import io

from .type import TType, RecursiveType

if TYPE_CHECKING:
    from .autospec import ThriftObject

    T = TypeVar('T', bound=ThriftObject)
else:
    T = TypeVar('T', bound='ThriftObject')

_alpha_start = ord("a")
_alpha_length = ord("z") - ord("a") + 1


class ThriftReader(io.BytesIO):
    """
    ThriftReader implements decoding the Thrift Compact protocol into Python values.

    https://github.com/apache/thrift/blob/master/doc/specs/thrift-compact-protocol.md
    """

    _prev_field_id: int
    _stack: List[int]

    _prev_struct_id: int

    @property
    def _struct_id(self) -> str:
        """An incrementing alphabetical identifier used for pretty-printing structs."""
        self._prev_struct_id += 1
        return (chr(_alpha_start + (self._prev_struct_id // _alpha_length))
                + chr(_alpha_start + self._prev_struct_id % _alpha_length))

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._prev_field_id = 0
        self._prev_struct_id = -1
        self._stack = []

    def _push_stack(self) -> None:
        self._stack.append(self._prev_field_id)
        self._prev_field_id = 0

    def _pop_stack(self) -> None:
        if self._stack:
            self._prev_field_id = self._stack.pop()

    def _read_byte(self, signed: bool = False) -> int:
        return int.from_bytes(self.read(1), "big", signed=signed)

    def reset(self) -> None:
        """Reset the parser to the start of the data."""
        self.seek(0)
        self._prev_field_id = 0
        self._prev_struct_id = -1
        self._stack = []

    @staticmethod
    def _from_zigzag(val: int) -> int:
        return (val >> 1) ^ -(val & 1)

    def read_int(self) -> int:
        return self._from_zigzag(self.read_varint())

    def read_varint(self) -> int:
        shift = 0
        result = 0
        while True:
            byte = self._read_byte()
            result |= (byte & 0x7f) << shift
            if (byte & 0x80) == 0:
                break
            shift += 7
        return result

    def read_field(self) -> Tuple[TType, int]:
        byte = self._read_byte()
        type = TType(byte & 0x0f)
        if type == TType.STOP:
            return type, -1
        field_id_delta = byte >> 4
        self._prev_field_id = (self.read_int() if field_id_delta == 0
                               else self._prev_field_id + field_id_delta)
        return type, self._prev_field_id

    def read_val(self, type: TType) -> Any:
        """
        Read a primitive value.

        Args:
            type: The type of value to read.

        Returns:

        """
        if type == TType.TRUE:
            return True
        elif type == TType.FALSE:
            return False
        elif type == TType.BYTE:
            return self._read_byte()
        elif type == TType.BINARY:
            return self.read(self.read_varint())
        elif type in (TType.I16, TType.I32, TType.I64):
            # All sizes of ints are decoded the same way from zigzag.
            return self.read_int()
        elif type == TType.DOUBLE:
            # Doubles are encoded as little endian
            # https://github.com/apache/thrift/blob/master/doc/specs/thrift-compact-protocol.md#double-encoding
            return struct.unpack("<d", self.read(8))
        else:
            raise ValueError(f"{type.name} is not a primitive type")

    def read_list_header(self) -> Tuple[TType, int]:
        """
        Read the type and length metadata of a list or set.

        https://github.com/apache/thrift/blob/master/doc/specs/thrift-compact-protocol.md#list-and-set

        Returns:
            A tuple containing the item type and length of the list or set.
        """
        header_byte = self._read_byte()

        # The upstream Thrift spec uses different element type identifiers for list and map types,
        # but Facebook just uses the same types as structs.
        item_type = TType(header_byte & 0x0f)

        length = header_byte >> 4
        if length == 0x0f:
            length = self.read_varint()

        return item_type, length

    def read_map_header(self) -> Tuple[TType, TType, int]:
        """
        Read the type and length metadata of a map.

        https://github.com/apache/thrift/blob/master/doc/specs/thrift-compact-protocol.md#map

        Returns:
            A tuple containing the key type, value type and length of the map.
        """
        pos = self.tell()
        if self._read_byte() == 0:
            # If the first byte is zero, the map is empty.
            return TType.STOP, TType.STOP, 0
        # Go back one byte so we can read the length varint normally.
        self.seek(pos)
        length = self.read_varint()

        types = self._read_byte()
        key_type = TType(types >> 4)
        value_type = TType(types & 0x0f)

        return key_type, value_type, length

    def skip(self, type: TType) -> None:
        """
        Skip the next field in the data. If the type is a struct, list, set or map, this will
        recursively skip everything it contains.

        Args:
            type: The type of the field.
        """
        if type == TType.STRUCT:
            self._push_stack()
            while True:
                field_type, _ = self.read_field()
                if field_type == TType.STOP:
                    break
                self.skip(field_type)
            self._pop_stack()
        elif type in (TType.LIST, TType.SET):
            item_type, length = self.read_list_header()
            for _ in range(length):
                self.skip(item_type)
        elif type == TType.MAP:
            key_type, value_type, length = self.read_map_header()
            for _ in range(length):
                self.skip(key_type)
                self.skip(value_type)
        else:
            self.read_val(type)

    def _read_kv(self, key_type: RecursiveType, value_type: RecursiveType, field_path: str,
                 index: int) -> Tuple[Any, Any]:
        key = self.read_val_recursive(key_type, field_path=f"{field_path}[{index}::key]")
        value_path = repr(key) if isinstance(key, (str, bytes, int)) else f"{index}::value"
        value = self.read_val_recursive(value_type, field_path=f"{field_path}[{value_path}]")
        return key, value

    def read_val_recursive(self, rtype: RecursiveType, field_path: str = "root") -> Any:
        """
        Read any type of value from the buffer.

        Args:
            rtype: The exact type specification for the value to read.
            field_path: The recursive field path used for debugging.

        Returns:
            The parsed value.
        """
        if rtype.type == TType.STRUCT:
            self._push_stack()
            val = self.read_struct(rtype.python_type, field_path=field_path)
            self._pop_stack()
            return val
        elif rtype.type == TType.MAP:
            key_type, value_type, length = self.read_map_header()
            if length == 0:
                return {}
            elif key_type != rtype.key_type.type:
                raise ValueError(f"Unexpected key type at {field_path}: "
                                 f"expected {rtype.key_type.type.name}, got {key_type.name}")
            elif value_type != rtype.value_type.type:
                raise ValueError(f"Unexpected value type at {field_path}: "
                                 f"expected {rtype.value_type.type.name}, got {value_type.name}")
            return dict(self._read_kv(rtype.key_type, rtype.value_type, field_path, index)
                        for index in range(length))
        elif rtype.type in (TType.LIST, TType.SET):
            item_type, length = self.read_list_header()
            if item_type != rtype.item_type.type:
                raise ValueError(f"Unexpected item type at {field_path}: "
                                 f"expected {rtype.item_type.type.name}, got {item_type.name}")
            data = (self.read_val_recursive(rtype.item_type, field_path=f"{field_path}[{i}]")
                    for i in range(length))
            return set(data) if rtype.type == TType.SET else list(data)
        else:
            if rtype.type == TType.BINARY and rtype.python_type != bytes:
                # For non-bytes python types, decode as UTF-8 and then call the
                # type constructor in case it's an enum or something like that.
                try:
                    return rtype.python_type(self.read_val(rtype.type).decode("utf-8"))
                except UnicodeDecodeError as e:
                    raise ValueError(f"Failed to decode string at {field_path}: {e}")
            return self.read_val(rtype.type)

    def read_struct(self, type: Type[T], field_path: str = "root") -> T:
        """
        Assuming the data in the buffer is a Thrift struct, parse it into a dataclass.

        Args:
            type: The Python type to parse the struct into.
            field_path: The recursive field path used for debugging.

        Returns:
            An instance of the given type with the parsed data.
        """
        args = {}
        while True:
            field_type, field_index = self.read_field()
            if field_type == TType.STOP:
                break
            try:
                field_meta = type.thrift_spec[field_index]
            except KeyError:
                # If the field isn't present in the class at all, ignore it.
                self.skip(field_type)
                continue
            expected_type = TType.BOOL if field_type in (TType.TRUE, TType.FALSE) else field_type
            if field_meta.rtype.type != expected_type:
                raise ValueError(f"Mismatching type for for field {field_meta.name}/#{field_index}"
                                 f": expected {field_meta.rtype.type.name}, got {field_type.name}")
            if expected_type == TType.BOOL:
                args[field_meta.name] = True if field_type == TType.TRUE else False
            else:
                fp = f"{field_path}.{field_meta.name}"
                args[field_meta.name] = self.read_val_recursive(field_meta.rtype, field_path=fp)
        try:
            return type(**args)
        except TypeError as e:
            raise ValueError(f"Failed to create {type.__name__} at {field_path}") from e

    def pretty_print(self, field_type: TType = TType.STRUCT, _indent: str = "", _prefix: str = ""
                     ) -> None:
        """
        Pretty-print the value in the reader.
        Useful for debugging and reverse-engineering schemas.
        """
        if _prefix:
            print(f"{_indent}{_prefix} ", end="")
        if field_type in (TType.LIST, TType.SET):
            item_type, length = self.read_list_header()
            print(f"{item_type.name} {length} items")
            for i in range(length):
                self.pretty_print(item_type, _indent + "  ", f"{i + 1}.")
        elif field_type == TType.MAP:
            key_type, value_type, length = self.read_map_header()
            print(f"<{key_type.name}: {value_type.name}> - {length} items")
            for _ in range(length):
                key = self.read_val(key_type)
                self.pretty_print(value_type, _indent + "  ", f"{key}:")
        elif field_type == TType.STRUCT:
            struct_id = self._struct_id
            print(f"start-{struct_id}")
            self._push_stack()
            while True:
                subfield_type, subfield_index = self.read_field()
                if subfield_type == TType.STOP:
                    break
                self.pretty_print(subfield_type, _indent + "  ",
                                  f"{subfield_index} ({subfield_type.name}):")
            print(f"{_indent}end-{struct_id}")
            self._pop_stack()
        else:
            print(self.read_val(field_type))



from alembic import context
from sqlalchemy import engine_from_config, pool
from logging.config import fileConfig

import sys
from os.path import abspath, dirname

sys.path.insert(0, dirname(dirname(abspath(__file__))))

from mautrix.util.db import Base
from mautrix_facebook.config import Config
import mautrix_facebook.db

config = context.config
mxfb_config_path = context.get_x_argument(as_dictionary=True).get("config", "config.yaml")
mxfb_config = Config(mxfb_config_path, None, None)
mxfb_config.load()
config.set_main_option("sqlalchemy.url", mxfb_config["appservice.database"].replace("%", "%%"))
fileConfig(config.config_file_name)
target_metadata = Base.metadata


def run_migrations_offline():
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url, target_metadata=target_metadata, literal_binds=True,
        render_as_batch=True)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix='sqlalchemy.',
        poolclass=pool.NullPool)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            render_as_batch=True,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Add bridge notice room column for users

Revision ID: a26da6af5cff
Revises: 97e9f7a3c070
Create Date: 2020-05-26 15:44:47.460965

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a26da6af5cff'
down_revision = '97e9f7a3c070'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("user") as batch_op:
        batch_op.add_column(sa.Column('notice_room', sa.String(length=255), nullable=True))


def downgrade():
    with op.batch_alter_table("user") as batch_op:
        batch_op.drop_column('notice_room')



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, TYPE_CHECKING, ClassVar

from asyncpg import Record
from attr import dataclass

from mautrix.types import RoomID, EventID
from mautrix.util.async_db import Database

fake_db = Database("") if TYPE_CHECKING else None


@dataclass
class Reaction:
    db: ClassVar[Database] = fake_db

    mxid: EventID
    mx_room: RoomID
    fb_msgid: str
    fb_receiver: int
    fb_sender: int
    reaction: str

    @classmethod
    def _from_row(cls, row: Optional[Record]) -> Optional['Reaction']:
        if row is None:
            return None
        return cls(**row)

    @classmethod
    async def get_by_fbid(cls, fb_msgid: str, fb_receiver: int, fb_sender: int
                          ) -> Optional['Reaction']:
        q = ("SELECT mxid, mx_room, fb_msgid, fb_receiver, fb_sender, reaction "
             "FROM reaction WHERE fb_msgid=$1 AND fb_receiver=$2 AND fb_sender=$3")
        row = await cls.db.fetchrow(q, fb_msgid, fb_receiver, fb_sender)
        return cls._from_row(row)

    @classmethod
    async def get_by_mxid(cls, mxid: EventID, mx_room: RoomID) -> Optional['Reaction']:
        q = ("SELECT mxid, mx_room, fb_msgid, fb_receiver, fb_sender, reaction "
             "FROM reaction WHERE mxid=$1 AND mx_room=$2")
        row = await cls.db.fetchrow(q, mxid, mx_room)
        return cls._from_row(row)

    async def insert(self) -> None:
        q = ("INSERT INTO reaction (mxid, mx_room, fb_msgid, fb_receiver, fb_sender, reaction) "
             "VALUES ($1, $2, $3, $4, $5, $6)")
        await self.db.execute(q, self.mxid, self.mx_room, self.fb_msgid, self.fb_receiver,
                              self.fb_sender, self.reaction)

    async def delete(self) -> None:
        q = "DELETE FROM reaction WHERE fb_msgid=$1 AND fb_receiver=$2 AND fb_sender=$3"
        await self.db.execute(q, self.fb_msgid, self.fb_receiver, self.fb_sender)

    async def save(self) -> None:
        q = ("UPDATE reaction SET mxid=$1, mx_room=$2, reaction=$3 "
             "WHERE fb_msgid=$4 AND fb_receiver=$5 AND fb_sender=$6")
        await self.db.execute(q, self.mxid, self.mx_room, self.reaction,
                              self.fb_msgid, self.fb_receiver, self.fb_sender)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union
import argparse

from sqlalchemy import orm
import sqlalchemy as sql

parser = argparse.ArgumentParser(description="mautrix-telegram dbms migration script",
                                 prog="python -m mautrix_telegram.scripts.dbms_migrate")
parser.add_argument("-f", "--from-url", type=str, required=True, metavar="<url>",
                    help="the old database path")
parser.add_argument("-t", "--to-url", type=str, required=True, metavar="<url>",
                    help="the new database path")
parser.add_argument("-s", "--silent", action="store_true", help="No logs while migrating")
args = parser.parse_args()
silent = args.silent or False


def log(message, end="\n"):
    if not silent:
        print(message, end=end, flush=True)


def connect(to):
    from mautrix.util.db import Base
    from mautrix.client.state_store.sqlalchemy import RoomState, UserProfile
    from .tables import Portal, Message, Puppet, User, UserPortal, Contact, Reaction

    db_engine = sql.create_engine(to)
    db_factory = orm.sessionmaker(bind=db_engine)
    db_session: Union[orm.Session, orm.scoped_session] = orm.scoped_session(db_factory)
    Base.metadata.bind = db_engine

    return db_session, {
        "Portal": Portal,
        "Message": Message,
        "Puppet": Puppet,
        "User": User,
        "RoomState": RoomState,
        "UserProfile": UserProfile,
        "Reaction": Reaction,
    }


log("Connecting to old database")
session, tables = connect(args.from_url)

data = {}
for name, table in tables.items():
    log(f"Reading table {name}...", end=" ")
    data[name] = session.query(table).all()
    log("Done!")

log("Connecting to new database")
session, tables = connect(args.to_url)

for name, table in tables.items():
    log(f"Writing table {name}", end="")
    length = len(data[name])
    n = 0
    for row in data[name]:
        session.merge(row)
        n += 5
        if n >= length:
            log(".", end="")
            n = 0
    log(" Done!")

log("Committing changes to database...", end=" ")
session.commit()
log("Done!")



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, Any, Optional


class ResponseError(Exception):
    def __init__(self, data: Dict[str, Any]) -> None:
        self.data = data
        user_message = data.get("error_user_msg")
        if user_message:
            super().__init__(user_message)
        else:
            message = data["message"]
            code = data["code"]
            subcode = data.get("subcode")
            code_str = f"{code}.{subcode}" if subcode else str(code)
            super().__init__(f"{code_str}: {message}")


class OAuthException(ResponseError):
    pass


class InvalidAccessToken(OAuthException):
    pass


class TwoFactorRequired(OAuthException):
    user_message: str
    login_first_factor: str
    auth_token: str
    machine_id: Optional[str]
    uid: int

    def __init__(self, data: Dict[str, Any]) -> None:
        super().__init__(data)
        tfa_data = data["error_data"]
        self.login_first_factor = tfa_data["login_first_factor"]
        self.machine_id = tfa_data.get("machine_id")
        self.auth_token = tfa_data["auth_token"]
        self.uid = tfa_data["uid"]


class InvalidEmail(OAuthException):
    pass


class IncorrectPassword(OAuthException):
    pass


class GraphMethodException(ResponseError):
    pass


error_code_map = {
    190: InvalidAccessToken,
    400: InvalidEmail,
    401: IncorrectPassword,
    406: TwoFactorRequired,
}
_error_classes = (OAuthException, GraphMethodException)
error_class_map = {clazz.__name__: clazz for clazz in _error_classes}



from typing import TYPE_CHECKING

from mautrix.bridge.commands import CommandEvent as BaseCommandEvent

if TYPE_CHECKING:
    from ..__main__ import MessengerBridge
    from ..user import User


class CommandEvent(BaseCommandEvent):
    bridge: 'MessengerBridge'
    sender: 'User'



"""Add profile set booleans to puppet table

Revision ID: f91274813e8c
Revises: 4face019555a
Create Date: 2020-11-10 02:11:41.981068

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'f91274813e8c'
down_revision = '4face019555a'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('puppet', schema=None) as batch_op:
        batch_op.add_column(sa.Column('avatar_set', sa.Boolean(), server_default=sa.false(), nullable=False))
        batch_op.add_column(sa.Column('name_set', sa.Boolean(), server_default=sa.false(), nullable=False))

    # Fill the database with assumed data so the bridge doesn't spam a ton of no-op profile updates
    op.execute("UPDATE puppet SET name_set=true WHERE name<>''")
    op.execute("UPDATE puppet SET avatar_set=true WHERE photo_id<>''")


def downgrade():
    with op.batch_alter_table('puppet', schema=None) as batch_op:
        batch_op.drop_column('name_set')
        batch_op.drop_column('avatar_set')
